// -----------------------------------------------------------------------------
//
// Copyright 2012(c) Analog Devices, Inc.
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//  - Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  - Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//  - Neither the name of Analog Devices, Inc. nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//  - The use of this software may or may not infringe the patent rights
//    of one or more patent holders.  This license does not release you
//    from the requirement that you obtain separate licenses from these
//    patent holders to use this software.
//  - Use of the software either in source or binary form, must be run
//    on or directly connected to an Analog Devices Inc. component.
//
// THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// INTELLECTUAL PROPERTY RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// -----------------------------------------------------------------------------
// FILE NAME : AD7450A.v
// MODULE NAME : AD7450A
// AUTHOR : acostina
// AUTHOR'S EMAIL : adrian.costina@analog.com
// -----------------------------------------------------------------------------
// SVN REVISION: 1553
// -----------------------------------------------------------------------------
// KEYWORDS : AD7450A
// -----------------------------------------------------------------------------
// PURPOSE : Driver for the AD7450A  12-Bit, 1 MSPS ADC
// -----------------------------------------------------------------------------
// REUSE ISSUES
// Reset Strategy      : Active low reset signal
// Clock Domains       : 100 MHz system clock
//                       External 18 MHz ADC sclk
// Critical Timing     : N/A
// Test Features       : N/A
// Asynchronous I/F    : N/A
// Instantiations      : N/A
// Synthesizable (y/n) : Y
// Target Device       : AD7450A
// Other               : The driver is intended to be used for AD7450A ADCs
// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

`timescale 1 ns / 1 ps //Use a timescale that is best for simulation.

//------------------------------------------------------------------------------
//----------- Module Declaration -----------------------------------------------
//------------------------------------------------------------------------------

module AD7450A

//----------- Ports Declarations -----------------------------------------------
(
    //clock and reset signals
    input               fpga_clk_i,     // system clock
    input               adc_clk_i,      // clock to be applied to ADC to read the conversions results
    input               reset_n_i,      // active low reset signal

    //IP control and data interface
    output reg  [15:0]  data_o,         // data read from the ADC
    output reg          data_rd_ready_o,// when set to high the data read from the ADC is available on the data_o bus

    //ADC control and data interface
    input               adc_sdata_i,    // ADC SDO signal
    output reg          adc_cs_n_o      // ADC CNV signal

);

//------------------------------------------------------------------------------
//----------- Registers Declarations -------------------------------------------
//------------------------------------------------------------------------------
reg [2:0]   presentState;       // Current FSM State
reg [2:0]   nextState;          // Next FSM State

reg [31:0]  adc_tcycle_cnt;     // counts the number of FPGA clock cycles to determine when an ADC cycle is complete
reg [ 5:0]  sclk_clk_cnt;       // counts the number of clocks applied to the ADC to read the conversion result

reg         data_rd_s;
reg         sync_data_rd;
reg         sync_data_rd_d1;
reg         flag;

reg [15:0]  data_o_s;           // data output internal signal

//------------------------------------------------------------------------------
//----------- Local Parameters -------------------------------------------------
//------------------------------------------------------------------------------

localparam  ADC_IDLE_STATE          = 3'b001;
localparam  ADC_ACQUIRE_DATA_STATE  = 3'b010;
localparam  ADC_WAIT_END_STATE      = 3'b100;

// ADC timing
parameter real FPGA_CLOCK_FREQ      = 18;  // FPGA clock frequency [MHz]
parameter real ADC_CYCLE_TIME       = 1;   // minimum time between two ADC conversions (Tcyc) [us]
parameter [31:0] ADC_CYCLE_CNT      = FPGA_CLOCK_FREQ * ADC_CYCLE_TIME - 3;

// ADC serial clock periods
parameter ADC_SCLK_PERIODS          = 6'd16;// number of clocks to be sent to the ADC to read the conversion result

//------------------------------------------------------------------------------
//----------- Assign/Always Blocks ---------------------------------------------
//------------------------------------------------------------------------------

// Count Cycle Time
always @(posedge adc_clk_i)
begin
    if( presentState == ADC_IDLE_STATE )
    begin
        adc_tcycle_cnt  <= ADC_CYCLE_CNT;
    end
    else if ( adc_tcycle_cnt > 0 )
    begin
        adc_tcycle_cnt <= adc_tcycle_cnt - 1;
    end
end

// Acquire data
always @(negedge adc_clk_i)
begin
    if( adc_cs_n_o == 1'b0 )
    begin
        data_o_s        <= { data_o_s[14:0], adc_sdata_i };
    end
end

// Register States
always @(posedge adc_clk_i)
begin
    presentState <= nextState;
end

// State switching logic
always @(*)
begin
    nextState   <= presentState;
    if( adc_tcycle_cnt == 0 )
    begin
        nextState <= ADC_IDLE_STATE;
    end
    else
    begin
        case(presentState)
            ADC_IDLE_STATE:
                begin
                    nextState <= ADC_ACQUIRE_DATA_STATE;
                end
            ADC_ACQUIRE_DATA_STATE:
                begin
                    if ( sclk_clk_cnt == 0 )
                    begin
                        nextState <= ADC_WAIT_END_STATE;
                    end
                end
            ADC_WAIT_END_STATE:
                begin
                    if( adc_tcycle_cnt == 0 )
                    begin
                        nextState <= ADC_IDLE_STATE;
                    end
                end
            default:
                begin
                    nextState   <= ADC_IDLE_STATE;
                end
        endcase
    end
end

// State output logic
always @(posedge adc_clk_i)
begin
    case( presentState )
        ADC_IDLE_STATE:
        begin
            adc_cs_n_o      <= 1'b1;
            sclk_clk_cnt    <= ADC_SCLK_PERIODS;
            data_rd_s       <= 1'b1;
            data_o          <= data_o_s;
        end
        ADC_ACQUIRE_DATA_STATE:
        begin
            adc_cs_n_o      <= 1'b0;
            sclk_clk_cnt    <= sclk_clk_cnt - 1'b1;
            data_rd_s       <= 1'b0;
        end
        ADC_WAIT_END_STATE:
        begin
            adc_cs_n_o      <= 1'b1;
            sclk_clk_cnt    <= ADC_SCLK_PERIODS;
            data_o          <= data_o_s;
            data_rd_s       <= 1'b0;
        end
        default:
        begin
            adc_cs_n_o      <= 1'b1;
            sclk_clk_cnt    <= ADC_SCLK_PERIODS;
        end
    endcase
end

// synchronize with the FPGA clock
always @(posedge fpga_clk_i)
begin
    sync_data_rd    <= data_rd_s;
    sync_data_rd_d1 <= sync_data_rd;
    data_rd_ready_o <= 1'b0;
    if (sync_data_rd_d1 == 1'b1 && flag == 1'b0)
    begin
        data_rd_ready_o <= 1'b1;
        flag            <= 1'b1;
    end
    else if (sync_data_rd_d1 == 1'b0)
    begin
        flag            <= 1'b0;
    end
end


endmodule
